from collections import defaultdict
from pathlib import Path
from typing import DefaultDict, Dict, Generator, List, Set, Union

import spacy
import srsly
from spacy.language import Language
from spacy.scorer import Scorer
from tqdm import tqdm, tqdm_notebook

from .constants import NONE
from .recognizer import EntityRecognizer
from .types import Example, PredictionError


def ents_by_label(
    data: List[Example], use_lower: bool = True
) -> DefaultDict[str, List[str]]:
    """Get a dictionary of unique text spans by label for your data
    
    ### Parameters
    --------------
    **data**: (List[Example]), required.
        List of Examples
    **use_lower**: (bool, optional), Defaults to True.
        Use the lowercase form of the span text
    
    ### Returns
    -----------
    (DefaultDict[str, List[str]]):
        DefaultDict mapping label to sorted list of the unique
        spans annotated for that label.
    """
    annotations: DefaultDict[str, Set[str]] = defaultdict(set)
    sorted_annotations: DefaultDict[str, List[str]] = defaultdict(list)

    for e in data:
        for s in e.spans:
            span_text = s.text.lower() if use_lower else s.text
            annotations[s.label].add(span_text)

    for label in annotations.keys():
        sorted_annotations[label] = sorted(annotations[label])

    return sorted_annotations


def get_label_disparities(
    data: List[Example], label1: str, label2: str, use_lower: bool = True
) -> Set[str]:
    """Identify annotated spans that have different labels in different examples
    
    ### Parameters
    --------------
    **data**: (List[Example]), required.
        Input List of Examples
    **label1**: (str), required.
        First label to compare
    **label2**: (str), required.
        Second label to compare
    **use_lower**: (bool, optional), Defaults to True.
        Use the lowercase form of the span text in ents_to_label
    
    ### Returns
    -----------
    (Set[str]): 
        Set of all unique text spans that overlap between label1 and label2
    """
    annotations = ents_by_label(data, use_lower=use_lower)
    return set(annotations[label1]).intersection(set(annotations[label2]))


def top_label_disparities(
    data: List[Example], use_lower: bool = True
) -> List[Dict[str, object]]:
    """Identify annotated spans that have different labels
    in different examples for all label pairs in data.
    
    ### Parameters
    --------------
    **data**: (List[Example]), required.
        Input List of Examples
    **use_lower**: (bool, optional), Defaults to True.
        Use the lowercase form of the span text in ents_to_label
        
    
    ### Returns
    -----------
    (): 
        [description]
    """
    annotations = ents_by_label(data, use_lower=use_lower)
    label_disparities = {}
    for label1 in annotations.keys():
        for label2 in annotations.keys():
            if label1 != label2:
                n_disparities = len(
                    set(annotations[label1]).intersection(set(annotations[label2]))
                )
                if n_disparities > 0:
                    input_hash = "||".join(sorted([label1, label2]))
                    label_disparities[input_hash] = {
                        "label1": label1,
                        "label2": label2,
                        "n_disparities": n_disparities,
                    }

    return sorted(
        label_disparities.values(), key=lambda row: row["n_disparities"], reverse=True
    )


def top_prediction_errors(
    ner: EntityRecognizer,
    data: List[Example],
    labels: List[str] = None,
    n: int = None,
    k: int = None,
    exclude_fp: bool = False,
    exclude_fn: bool = False,
    verbose: bool = False,
) -> List[PredictionError]:
    """Get a sorted list of examples your model is worst at predicting.
    
    ### Parameters
    --------------
    **ner**: (EntityRecognizer), required.
        An instance of EntityRecognizer
    **data**: (List[Example]), required.
        List of annotated Examples
    **labels**: (List[str], optional), Defaults to None.
        List of labels to get errors for. Defaults to the labels property of `ner`.
    **n**: (int, optional), Defaults to None.
        If set, only use the top n examples from data
    **k**: (int, optional), Defaults to None.
        If set, return the top k prediction errors, otherwise the whole list.
    **exclude_fp**: (bool, optional), Defaults to False.
        Flag to exclude False Positive errors.
    **exclude_fn**: (bool, optional), Defaults to False.
        Flag to exclude False Negative errors.
    **verbose**: (bool, optional), Defaults to False.
        Show progress_bar or not
    
    ### Returns
    -----------
    (List[PredictionError]): 
        [description]
    """

    labels_ = labels or ner.labels
    if n is not None:
        data = data[:n]

    n_examples = len(data)
    texts = (e.text for e in data)
    anns = (e.spans for e in data)

    errors = defaultdict(lambda: defaultdict(lambda: defaultdict(int)))  # type: ignore
    error_examples: DefaultDict[str, List[Example]] = defaultdict(list)

    for orig_example, pred, ann in zip(data, ner.predict(texts), anns):

        cand = set([(s.start, s.end, s.label) for s in pred.spans])
        gold = set([(s.start, s.end, s.label) for s in ann])

        fp_diff = cand - gold
        fn_diff = gold - cand

        if fp_diff and not exclude_fp:
            for fp in fp_diff:
                for gold_ent in gold:
                    if fp[0] == gold_ent[0] and fp[1] == gold_ent[1]:
                        start, end, label = gold_ent
                        false_label = fp[2]
                        text = pred.text[start:end]
                        errors[label][text][false_label] += 1
                        error_examples[text].append(orig_example)

        if fn_diff and not exclude_fn:
            for fn in fn_diff:
                has_gold_ent = False
                for gold_ent in gold:
                    if fp[0] == gold_ent[0] and fp[1] == gold_ent[1]:
                        has_gold_ent = True

                if not has_gold_ent:
                    start, end, label = fn
                    text = pred.text[start:end]
                    errors[label][text][NONE] += 1
                    error_examples[text].append(orig_example)

    ranked_errors: List[PredictionError] = []

    for label, errors_per_label in errors.items():
        for error_text, error_labels in errors_per_label.items():
            for error_label, count in error_labels.items():
                ranked_errors.append(
                    PredictionError(
                        text=error_text,
                        true_label=label,
                        pred_label=error_label,
                        count=count,
                        examples=error_examples[error_text],
                    )
                )

    ranked_errors = sorted(ranked_errors, key=lambda error: error.count, reverse=True)
    if k:
        ranked_errors = ranked_errors[:k]
    return ranked_errors



from pathlib import Path
from typing import Dict, List

import pytest
from recon.loaders import read_jsonl
from recon.stats import ner_stats
from recon.types import Example
from spacy.lang.en import English


@pytest.fixture()
def nlp():
    return English()


@pytest.fixture()
def test_texts():
    return [
        "Machine learning is the most researched area of AI.",
        "My title at work is Software Engineer even theough I mostly work on AI.",
    ]


@pytest.fixture()
def example_data() -> Dict[str, List[Example]]:
    """Fixture to load example train/dev/test data that has inconsistencies.
    
    ### Returns
    -----------
    (Dict[str, List[Example]]):
        Dataset containing the train/dev/test split
    """
    base_path = Path(__file__).parent.parent / "examples/data"
    return {
        "train": read_jsonl(base_path / "train.jsonl"),
        "dev": read_jsonl(base_path / "dev.jsonl"),
        "test": read_jsonl(base_path / "test.jsonl"),
    }



from typing import Any, Dict, List, Optional

from pydantic import BaseModel, Field, Schema, validator


class TextSpanLabel(BaseModel):
    text: str
    start: int
    end: int
    label: str


class Example(BaseModel):
    text: str
    spans: List[TextSpanLabel]
    meta: Dict[str, Any] = {}


class PredictionError(BaseModel):
    text: str
    true_label: str
    pred_label: str
    count: int
    examples: Optional[List[Example]] = []



from recon.stats import ner_stats


def test_ner_stats(example_data):
    stats = ner_stats(example_data["train"])
    printed_stats = ner_stats(example_data["train"], serialize=True)
    str_stats = ner_stats(example_data["train"], serialize=True, no_print=True)
    assert printed_stats is None
    assert isinstance(str_stats, str)

    assert stats["n_examples"] == 102
    assert stats["n_examples_no_entities"] == 29
    assert stats["ents_per_type"] == {"SKILL": 191, "PRODUCT": 34, "JOB_ROLE": 5}



import pytest
from recon.recognizer import *


def test_base_recognizer(test_texts):
    recognizer = EntityRecognizer()

    with pytest.raises(NotImplementedError):
        recognizer.labels
    with pytest.raises(NotImplementedError):
        recognizer.predict(test_texts)


def test_spacy_recognizer(nlp, test_texts):
    ruler = nlp.create_pipe("entity_ruler")
    ruler.add_patterns(
        [
            {"label": "SKILL", "pattern": "Machine learning"},
            {"label": "SKILL", "pattern": "researched"},
            {"label": "SKILL", "pattern": "AI"},
            {"label": "JOB_ROLE", "pattern": "Software Engineer"},
        ]
    )

    nlp.add_pipe(ruler)

    recognizer = SpacyEntityRecognizer(nlp)
    assert recognizer.labels == ["JOB_ROLE", "SKILL"]

    examples = list(recognizer.predict(test_texts))

    assert examples[0].text == test_texts[0]
    assert len(examples[0].spans) == 3
    assert len(examples[1].spans) == 2



from typing import Iterable, Iterator, List, Set

from spacy.language import Language

from .types import Example, TextSpanLabel


class EntityRecognizer:
    @property
    def labels(self) -> List[str]:
        """Return List of String Labels
        
        ### Raises
        -----------
        NotImplementedError: 
            Not Implemented, override
        
        ### Returns
        -----------
        (List[str]): 
            List of labels the model can predict
        """
        raise NotImplementedError

    def predict(self, texts: Iterable[str]) -> Iterator[Example]:
        """Run model inference on a batch of raw texts.
        
        ### Parameters
        --------------
        **texts**: (List[str]), required.
            Raw text examples
        
        ### Raises
        -----------
        NotImplementedError: 
            Not implemented, override
        
        ### Returns
        -----------
        (Iterator[Example]): 
            Examples constructed from Model predictions
        """
        raise NotImplementedError


class SpacyEntityRecognizer(EntityRecognizer):
    def __init__(self, nlp: Language):
        super().__init__()
        self.nlp = nlp

    @property
    def labels(self) -> List[str]:
        """Return List of spaCy ner labels
        
        ### Returns
        -----------
        (List[str]): 
            List of labels from spaCy ner pipe
        """
        all_labels: Set[str] = set()

        for pipe in ["ner", "entity_ruler"]:
            if self.nlp.has_pipe(pipe):
                all_labels = all_labels | set(self.nlp.get_pipe(pipe).labels)

        return sorted(list(all_labels))

    def predict(self, texts: Iterable[str]) -> Iterator[Example]:
        """Run spaCy nlp.pipe on a batch of raw texts.
        
        ### Parameters
        --------------
        **texts**: (List[str]), required.
            Raw text examples
        
        ### yields
        ----------
        (Iterator[Example]): 
            Examples constructed from spaCy Model predictions
        """

        examples: List[Example] = []

        for doc in self.nlp.pipe(texts):
            yield Example(
                text=doc.text,
                spans=[
                    TextSpanLabel(
                        text=e.text, start=e.start_char, end=e.end_char, label=e.label_
                    )
                    for e in doc.ents
                ],
            )



import pytest
from recon.types import Example, TextSpanLabel
from recon.validation import *


@pytest.fixture()
def messy_data():
    return [
        {
            "text": "Denver, Colorado is a city.",
            "spans": [
                {"start": 0, "end": 6, "label": "GPE"},
                {"start": 0, "end": 16, "label": "LOC"},
            ],
            "meta": "Cities Data",
        }
    ]


def test_json_to_examples(messy_data):
    fixed_examples = json_to_examples(messy_data)
    assert isinstance(fixed_examples[0], Example)
    assert isinstance(fixed_examples[0].spans[0], TextSpanLabel)

    assert fixed_examples[0].spans[0].text == "Denver, Colorado"
    assert fixed_examples[0].spans[0].label == "LOC"


def test_fix_annotations_format(messy_data):
    fixed_data = fix_annotations_format(messy_data)

    assert fixed_data[0]["spans"][0]["text"] == "Denver"
    assert fixed_data[0]["meta"] == {"source": "Cities Data"}


# def test_remove_overlaps():
#     test_entities = [(0, 5, "ENTITY"), (6, 10, "ENTITY")]
#     result = remove_overlapping_entities(test_entities)
#     assert result == [(0, 5, "ENTITY"), (6, 10, "ENTITY")]

#     test_entities = [(0, 5, "ENTITY"), (5, 10, "ENTITY")]
#     result = remove_overlapping_entities(test_entities)
#     assert result == [(0, 5, "ENTITY"), (5, 10, "ENTITY")]

#     test_entities = [(0, 5, "ENTITY"), (4, 10, "ENTITY")]
#     result = remove_overlapping_entities(test_entities)
#     assert result == [(4, 10, "ENTITY")]

#     test_entities = [(0, 5, "ENTITY"), (0, 5, "ENTITY")]
#     result = remove_overlapping_entities(test_entities)
#     assert result == [(0, 5, "ENTITY")]

#     test_entities = [(0, 5, "ENTITY"), (4, 11, "ENTITY"), (6, 20, "ENTITY")]
#     result = remove_overlapping_entities(test_entities)
#     assert result == [(0, 5, "ENTITY"), (6, 20, "ENTITY")]

#     test_entities = [(0, 5, "ENTITY"), (4, 7, "ENTITY"), (10, 20, "ENTITY")]
#     result = remove_overlapping_entities(test_entities)
#     assert result == [(0, 5, "ENTITY"), (10, 20, "ENTITY")]

#     test_entities = [(1368, 1374, "ENTITY"), (1368, 1376, "ENTITY")]
#     result = remove_overlapping_entities(test_entities)
#     assert result == [(1368, 1376, "ENTITY")]

#     test_entities = [
#         (12, 33, "ENTITY"),
#         (769, 779, "ENTITY"),
#         (769, 787, "ENTITY"),
#         (806, 811, "ENTITY"),
#     ]
#     result = remove_overlapping_entities(test_entities)
#     assert result == [(12, 33, "ENTITY"), (769, 787, "ENTITY"), (806, 811, "ENTITY")]

#     test_entities = [
#         (189, 209, "ENTITY"),
#         (317, 362, "ENTITY"),
#         (345, 354, "ENTITY"),
#         (364, 368, "ENTITY"),
#     ]
#     result = remove_overlapping_entities(test_entities)
#     assert result == [(189, 209, "ENTITY"), (317, 362, "ENTITY"), (364, 368, "ENTITY")]

#     test_entities = [(445, 502, "ENTITY"), (461, 473, "ENTITY"), (474, 489, "ENTITY")]
#     result = remove_overlapping_entities(test_entities)
#     assert result == [(445, 502, "ENTITY")]



from pathlib import Path
from typing import Any, Callable, Dict, List

from spacy.util import ensure_path

from .loaders import read_json, read_jsonl
from .types import Example


class Dataset:
    """Container for a full dataset with train/dev/test splits.
    Used to apply core functions to all datasets at once.

    ### Parameters
    --------------
    **train**: (List[Example]), required.
        List of Examples for **train** set
    **dev**: (List[Example]), required.
        List of Examples for **dev** set
    **test**: (List[Example], optional), Defaults to None.
        List of Examples for **test** set
    """

    def __init__(
        self, train: List[Example], dev: List[Example], test: List[Example] = None
    ):
        self._train = train
        self._dev = dev
        self._test = test

    @classmethod
    def from_disk(cls, path: Path, loader_func: Callable = read_jsonl) -> "Dataset":
        """Load Dataset from disk given a directory with files 
        named explicitly train.jsonl, dev.jsonl, and test.jsonl
        
        ### Parameters
        --------------
        **path**: (Path), required.
            directory to load from
        **loader_func**: (Callable, optional), Defaults to read_jsonl.
            Loader function (TODO: Make this a bit more generic)
        """

        path = ensure_path(path)
        return Dataset(
            loader_func(path / "train.jsonl"),
            loader_func(path / "dev.jsonl"),
            test=loader_func(path / "test.jsonl"),
        )

    @property
    def train(self) -> List[Example]:
        return self._train

    @property
    def dev(self) -> List[Example]:
        return self._dev

    @property
    def test(self) -> List[Example]:
        return self._test or []

    @property
    def all(self) -> List[Example]:
        return self.train + self.dev + self.test

    def apply(
        self, func: Callable[[List[Example], Any, Any], Any], *args: Any, **kwargs: Any
    ) -> Dict[str, Any]:
        """Apply an existing function to all datasets
        
        ### Parameters
        --------------
        **func**: (Callable[[List[Example]], Any]), required.
            Function from an existing recon module that can operate on a List of Examples
        
        ### Returns
        -----------
        (Dict[str, Any): 
            Dictionary mapping dataset names to return type of func Callable
        """
        res = {
            "train": func(self.train, *args, **kwargs),  # type: ignore
            "dev": func(self.dev, *args, **kwargs),  # type: ignore
            "test": func(self.test, *args, **kwargs),  # type: ignore
            "all": func(self.all, *args, **kwargs),  # type: ignore
        }

        return res

    def _validate_inplace_apply(
        self, old_data: List[Example], new_data: List[Example]
    ) -> List[Example]:
        assert isinstance(new_data, list)
        assert len(new_data) == len(old_data)
        assert isinstance(new_data[0], Example)
        return new_data

    def apply_(
        self,
        func: Callable[[List[Example], Any, Any], List[Example]],
        *args: Any,
        **kwargs: Any
    ) -> None:
        """Apply an existing function to all data inplace
        
        ### Parameters
        --------------
        **func**: (Callable[[List[Example]], List[Example]]), required.
            Function from an existing recon module that can operate on a List of Examples
        """

        new_train = self._validate_inplace_apply(self.train, func(self.train, *args, **kwargs))  # type: ignore
        new_dev = self._validate_inplace_apply(self.dev, func(self.dev, *args, **kwargs))  # type: ignore
        new_test = self._validate_inplace_apply(self.test, func(self.test, *args, **kwargs))  # type: ignore

        self._train = new_train
        self._dev = new_dev
        self._test = new_test



from pathlib import Path

import typer
from recon.dataset import Dataset
from recon.stats import ner_stats


def main(data_dir: Path):
    ds = Dataset.from_disk(data_dir)
    train_stats = ner_stats(ds.train)
    ner_stats(ds.train, serialize=True)


if __name__ == "__main__":
    typer.run(main)



from pathlib import Path

from recon.dataset import Dataset
from recon.stats import ner_stats


def test_dataset_initialize(example_data):
    ds1 = Dataset(example_data["train"], example_data["dev"])
    assert len(ds1.train) == 102
    assert len(ds1.dev) == 110
    assert len(ds1.all) == 212

    ds2 = Dataset(example_data["train"], example_data["dev"], test=example_data["test"])
    assert len(ds2.train) == 102
    assert len(ds2.dev) == 110
    assert len(ds2.test) == 96
    assert len(ds2.all) == 308


def test_dataset_from_disk(example_data):
    ds = Dataset.from_disk(Path(__file__).parent.parent / "examples/data")
    assert ds.train == example_data["train"]
    assert ds.dev == example_data["dev"]
    assert ds.test == example_data["test"]
    assert len(ds.all) == 308


def test_dataset_apply(example_data):
    dataset = Dataset(
        example_data["train"], example_data["dev"], test=example_data["test"]
    )
    stats = dataset.apply(ner_stats)
    assert sorted(list(stats.keys())) == ["all", "dev", "test", "train"]

    assert stats["all"]["n_examples"] == 308



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['recon']

package_data = \
{'': ['*']}

install_requires = \
['pydantic == 1.4', 'spacy >= 2.2.1, <3.0.0', 'typer >= 0.0.8', 'tqdm']

extras_require = \
{'all': ['colorama', 'click-completion'],
 'dev': ['jupyterlab >= 1.2.0, <1.3'],
 'doc': ['mkdocs', 'mkdocs-material', 'markdown-include', 'mkautodoc'],
 'test': ['pytest >=4.4.0',
          'pytest-cov',
          'coverage',
          'pytest-xdist',
          'pytest-sugar',
          'mypy',
          'black',
          'isort']}

setup(name='reconner',
      version='0.1.0',
      description='ReconNER, Debug annotated Named Entity Recognition (NER) data for inconsitencies and get insights on improving the quality of your data.',
      author='Kabir Khan',
      author_email='kakh@microsoft.com',
      url='https://github.com/microsoft/reconner',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=3.6',
     )



from collections import defaultdict
from typing import Any, DefaultDict, Dict, List, Optional, Union

import srsly

from .constants import NONE
from .types import Example, TextSpanLabel


def ner_stats(
    data: List[Example], serialize: bool = False, no_print: bool = False
) -> Union[List[Example], Dict[str, object], str, None]:
    """Compute statistics for NER data
    
    ### Parameters
    --------------
    **data**: (List[Example]), required.
        Data as a List of Examples
    **serialize**: (bool, optional), Defaults to False.
        Serialize to a JSON string for printing
    **no_print**: (bool, optional), Defaults to False.
        Don't print, return serialized string. Requires serialize to be True
    
    ### Returns
    -----------
    (List[Example]): 
        List of examples or string if serialize and no_print are both True
    """
    labels: DefaultDict[str, Any] = defaultdict(int)
    examples: DefaultDict[str, Any] = defaultdict(list)
    n_examples_no_entities = 0
    for e in data:
        if not e.spans:
            n_examples_no_entities += 1
            examples[NONE].append(e)
        else:
            for s in e.spans:
                labels[s.label] += 1
                examples[s.label].append(e)

    res = {
        "n_examples": len(data),
        "n_examples_no_entities": n_examples_no_entities,
        "ents_per_type": labels,
    }
    if serialize:
        serialized_res = srsly.json_dumps(res, indent=4)
        if no_print:
            return serialized_res
        else:
            print(serialized_res)
            return None
    else:
        res["examples_with_type"] = examples
        return res



import copy
from collections import defaultdict
from typing import DefaultDict, Dict, List

from .types import Example


def rename_labels(data: List[Example], label_map: Dict[str, str]) -> List[Example]:
    """Rename labels in a copy of List[Example] data
    
    ### Parameters
    --------------
    **data**: (List[Example]), required.
        List of Examples
    **label_map**: (Dict[str, str]), required.
        One-to-one mapping of label names
    
    ### Returns
    -----------
    (List[Example]): 
        Copy List of Examples with renamed labels
    """
    data_copy = copy.deepcopy(data)
    for example in data_copy:
        for span in example.spans:
            span.label = label_map.get(span.label, span.label)
    return data_copy


def fix_annotations(
    data: List[Example], corrections: Dict[str, str], use_lower: bool = True
) -> List[Example]:
    """Fix annotations in a copy of List[Example] data.
    This function will NOT add annotations to your data.
    It will only remove erroneous annotations and fix the
    labels for specific spans.
    
    ### Parameters
    --------------
    **data**: (List[Example]), required.
        List of Examples
    **corrections**: (Dict[str, str]), required.
        Dictionary of corrections mapping entity text to a new label.
        If the value is set to None, the annotation will be removed
    **use_lower**: (bool, optional), Defaults to True.
        Use the lowercase form of the span text for matching corrections
    
    ### Returns
    -----------
    (List[Example]): 
        Copy of list of Examples with corrections applied
    """
    data_copy: List[Example] = copy.deepcopy(data)
    if use_lower:
        corrections = {t.lower(): l for t, l in corrections.items()}

    prints: DefaultDict[str, List[str]] = defaultdict(list)

    for example in data:
        ents_to_remove = []
        for i, s in enumerate(example.spans):
            t = s.text.lower()

            if t in corrections:
                if corrections[t] is print:
                    prints[t] += [("=" * 100), example.text, s.label]
                elif corrections[t] is None:
                    ents_to_remove.append(i)
                else:
                    s.label = corrections[t]

        i = len(ents_to_remove) - 1
        while i >= 0:
            idx = ents_to_remove[i]
            del example.spans[idx]
            i -= 1

    for k in sorted(prints):
        print(f"**{k}**")
        for line in prints[k]:
            print(line)

    return data
